/*
Star Fox 2 LZ Big-Endian Encoder
*/

#include <stdio.h>
#include <stdlib.h>
#include <memory.h>

#include <vector>
#include <deque>

using namespace std;

#define u32 unsigned int
#define u8  unsigned char
#define s8    signed char

typedef struct
{
	int type;
	int pos;
	int ptr;
	int len;
} lz_find;

FILE *in,*out;

// LZ-related
vector<int> table[256];
deque<lz_find> lz_table;

int min_match = 2;
int window_size = 0xffff;
int max_match = 0xff;

// encoder-related
u8 buffer[0x10000];
int counter;
int dest = 0xffff;
int lz_ptr = 0;
bool first = 1;
int start;

// file-related
u8 file[0x10000];
int file_ptr = 0xffff;
u32 file_buf = 0;
int mask = 0;
int size = 0;

////////////////////////////////////////////////////
////////////////////////////////////////////////////

void WRITE_BYTE(int out, FILE *fp)
{
	file[file_ptr--] = out & 0xff;
}

void WRITE_WORD(int out, FILE *fp)
{
	WRITE_BYTE(out, fp);
	WRITE_BYTE(out>>8, fp);
}

void WRITE_DWORD(int out, FILE *fp)
{
	WRITE_WORD(out, fp);
	WRITE_WORD(out>>16, fp);
}

void PutBits(int bits, int data)
{
	// only use 'x' bits
	while(bits)
	{
		int attach;

		// append bits
		attach = data & (1<<(bits-1));
		attach >>= (bits-1);

		file_buf >>= 1;
		file_buf |= (attach<<31);

		// subtract usage
		mask--;
		bits--;

		// empty barrel
		if(!mask)
		{
			// compensate
			if(first)
			{
				file_buf >>= 1;
				file_buf |= (1<<31);
				first = 0;
				WRITE_DWORD(file_buf,out);
			}
			else
			{
				file_buf >>= 24;
				WRITE_BYTE(file_buf,out);
			}

			file_buf = 0;
			mask = 8;
		}
	}
}

///////////////////////////////////////////////////
///////////////////////////////////////////////////

void Find_LZ(int start, u8 byte, int &length, int &pos)
{
	int longest_length=0;
	int longest_ptr=0;

	// use cached positions to quickly look through the file
	for(int lcv=0; lcv<table[byte].size(); lcv++)
	{
		int match_length = 0;
		int ptr = table[byte][lcv];

		// invalid string; stop scanning
		if(ptr<=start)
			break;

		// LZ window restriction
		if(ptr-start>window_size)
			continue;

		// search for longest identical substring
		for(int lcv2=start; lcv2>=0 && match_length<max_match; lcv2--)
		{
			// look for a mismatch
			if(buffer[lcv2] != buffer[ptr])
				break;

			// keep looking
			ptr--;
			match_length++;
		}

		// record new long find
		if(longest_length <= match_length)
		{
			longest_ptr = table[byte][lcv];
			longest_length = match_length;
		}
	}

	// output findings
	length = longest_length;
	pos = longest_ptr;
}

///////////////////////////////////////////////////
///////////////////////////////////////////////////

void Encode_Free()
{
	lz_find lz = lz_table[lz_ptr++];

	// Select decoder method
	if(lz.type == 0)
	{
		if(lz.len <= 6)
		{
			// Print method (0-6)
			PutBits(3,lz.len);
		}
		else if(lz.len <= 15+7)
		{
			// Print method (7-22)
			PutBits(3,7);
			PutBits(1,0);
			PutBits(4,lz.len-7);
		}
		else if(lz.len <= 1023)
		{
			// Print method (23-1023)
			PutBits(3,7);
			PutBits(1,1);
			PutBits(10,lz.len);
		}
		else
		{
			// Print method (1024+)
			PutBits(3,7);
			PutBits(1,1);
			PutBits(10,0);
			PutBits(18,lz.len);
		}
	}
	// No data
	else
	{
		// Print method
		PutBits(3,0);
	}

	// Print free bytes
	while(lz.len--)
	{
		PutBits(8,buffer[start--]);
	}
}

void Encode_LZ()
{
	lz_find lz = lz_table[lz_ptr++];

	// Select decoder method
	if(lz.type == 1)
	{
		if(lz.len==2 && lz.ptr<=0xff)
		{
			// Print method
			PutBits(2,0);
			PutBits(8,lz.ptr);
		}
		else if(lz.len==3 && lz.ptr<=0x3fff)
		{
			// Print method
			PutBits(2,1);

			if(lz.ptr<=0xff)
			{
				PutBits(1,1);
				PutBits(8,lz.ptr);
			}
			else
			{
				PutBits(1,0);
				PutBits(14,lz.ptr);
			}
		}
		else // extended
		{
			// Print run
			if(lz.len == 4)
			{
				PutBits(2,2);
			}
			else if(5<=lz.len && lz.len<=6)
			{
				PutBits(2,3);
				PutBits(2,lz.len-5);
			}
			else if(7<=lz.len && lz.len<=10)
			{
				PutBits(2,3);
				PutBits(2,2);
				PutBits(2,lz.len-7);
			}
			else // lz.len <= 0xff
			{
				PutBits(2,3);
				PutBits(2,3);
				PutBits(8,lz.len);
			}

			// Print delta
			if(lz.ptr <= 0xff)
			{
				PutBits(1,1);
				PutBits(1,1);
				PutBits(8,lz.ptr);
			}
			else if(lz.ptr <= 0xfff)
			{
				PutBits(1,1);
				PutBits(1,0);
				PutBits(12,lz.ptr);
			}
			else // lz.ptr <= 0xffff
			{
				PutBits(1,0);
				PutBits(16,lz.ptr);
			}
		}
	}
	// No data
	else
	{
		// Print method
		PutBits(2,3);
		PutBits(2,3);
		PutBits(8,0);
	}

	start -= lz.len;
}

void Encode()
{
	lz_find lz;
	int counter=0;
	int free=0;

	// Step 1: Find all LZ matches
	start = dest;
	while(start>=0)
	{
		int future_length=0;
		int future_pos=0;

		int length;
		int pos;

		// Prepare for LZ
		table[ buffer[start] ].push_back(start);

		// Go find the longest substring match (and future ones)
		Find_LZ(start, buffer[start], length, pos);

		if(start-1 >= 0)
		{
			// Slightly increase ratio performance
			start--; table[ buffer[start] ].push_back(start);
			Find_LZ(start, buffer[start], future_length, future_pos);

			// Un-do lookahead
			table[ buffer[start] ].pop_back(); start++;
			if(future_length - length >= 1)
				length=0;
		}

		// Optimize further
		if(length==2 && (pos-start)>0xff)
			length=0;
		else if(length==3 && (pos-start)>0x3fff)
			length=0;

		if(length >= min_match)
		{
			// Flush free bytes first
			lz.type = 0;
			lz.len = free;

			lz_table.push_back(lz);

			// Found substring match; record and re-do
			lz.type = 1;
			lz.pos = start;
			lz.ptr = pos-start;
			lz.len = length;
			
			lz_table.push_back(lz);

			// Need to add to LZ table
			for(int lcv=1; lcv<length; lcv++)
				table[ buffer[start-lcv] ].push_back(start-lcv);

			// Fast update
			start -= length;
			free=0;
		}
		else
		{
			// No substrings found; try again
			start--;
			free++;
		}
	}

	// Flush free bytes again
	lz.type = 0;
	lz.len = free;

	lz_table.push_back(lz);

	// insert dummy entry
	lz.type = -1;
	lz_table.push_back(lz);

//////////////////////////////////////////////////
//////////////////////////////////////////////////

	// Init
	mask = 31;

	// Step 2a: Header
	WRITE_WORD(size,out);
	WRITE_WORD(0,out);
	
	// Step 2: Create method patterns
	start = dest;
	while(1)
	{
		Encode_Free();
		if(start >= 0) Encode_LZ();
		else break;
	}

	// Flush
	if(mask<8) PutBits(mask,0);
}

///////////////////////////////////////////////////
///////////////////////////////////////////////////

int main(int argc, char *argv[])
{
	if(argc<3) return 0;

	in = fopen(argv[1],"rb");

	// buffer file data (reverse storage)
	fread(buffer,0x10000,1,in);	
	size = ftell(in);
	dest = size-1;

	Encode();

	// Insert into image
	if(argc==4)
	{
		int offset;
		sscanf(argv[3], "%08X", &offset);
		out = fopen(argv[2],"r+b");
		int a = offset-(0x10000-(file_ptr+1));
		fseek(out,offset-(0x10000-(file_ptr+1)),SEEK_SET);
	}
	else
		out = fopen(argv[2],"wb");

	// write to file
	for(int lcv=file_ptr+1; lcv<0x10000; lcv++)
		fputc(file[lcv],out);

	fclose(in);
	fclose(out);

	return 0;
}
